/**
* TimrJS v3.1.1
* https://github.com/joe-castle/timrjs
* https://www.npmjs.com/package/timrjs
*
* Compatible with Browsers and Node.js (CommonJS).
*
* Copyright (c) 2021 Joe Castle
* Released under the MIT license
* https://github.com/joe-castle/timrjs/blob/master/LICENSE.md
*/
/**
 * Pads the beginning of a number that is greater than or equal 0 and less than 10 with a 0.
 *
 * @param {number} num the number to check
 *
 * @return {string|number} The padded number or the original argument.
 */
function zeroPad(num) {
    return (num >= 0 && num < 10) ? `0${num}` : num;
}

/**
 * Checks if a value is a valid number. Guards against NaN and Infinity.
 *
 * @param {any} value the value to be checked
 *
 * @return {boolean} True if it is, false if it isn't
 */
function isNum(value) {
    return typeof value === 'number' &&
        !isNaN(value) &&
        value !== Infinity &&
        value !== -Infinity;
}
/**
 * Checks if a value is a not a number.
 *
 * @param {any} value the value to be checked
 *
 * @return {boolean} True if it is, false if it isn't
 */
function isNotNum(value) {
    return !isNum(value);
}
/**
 * Checks if a value is a string.
 *
 * @param {any} value the value to be checked
 *
 * @return {boolean} True if it is, false if it isn't
 */
function isStr(value) {
    return typeof value === 'string';
}
/**
 * Checks if a value is not a string.
 *
 * @param {any} value the value to be checked
 *
 * @return {boolean} True if it is, false if it isn't
 */
function isNotStr(value) {
    return !isStr(value);
}
/**
 * Checks if a value is a boolean.
 *
 * @param {any} value the value to be checked
 *
 * @return {boolean} True if it is, false if it isn't
 */
function isBool(value) {
    return typeof value === 'boolean';
}
/**
 * Checks if a value is not a boolean.
 *
 * @param {any} value the value to be checked
 *
 * @return {boolean} True if it is, false if it isn't
 */
function isNotBool(value) {
    return !isBool(value);
}
/**
 * Checks if a value is a function.
 *
 * @param {any} value the value to be checked
 *
 * @return {boolean} True if it is, false if it isn't
 */
function isFn(value) {
    return typeof value === 'function';
}
/**
 * Checks if a value is not a function.
 *
 * @param {any} value the value to be checked
 *
 * @return {boolean} True if it is, false if it isn't
 */
function isNotFn(value) {
    return !isFn(value);
}
/**
 * Checks if a value is an object. Guards against null and arrays.
 *
 * @param {any} value the value to be checked
 *
 * @return {boolean} True if it is, false if it isn't
 */
function isObj(value) {
    return value !== null &&
        typeof value === 'object' &&
        !Array.isArray(value);
}
/**
 * Checks if the provided object extends the provided class
 *
 * @param {value} value the value to be checked
 * @param {any} clazz the class to check against
 *
 * @return True if it is, false if it isn't
 */
function isInstanceOf(value, clazz) {
    return value instanceof clazz;
}
/**
 * Checks if a value exists, i.e. not undefined or null.
 *
 * @param {any} value the value to be checked
 *
 * @return {boolean} True if it is, false if it isn't
 */
function exists(value) {
    return value !== undefined && value !== null;
}
/**
 * Checks if a value doesn't exist, i.e is undefined.
 *
 * @param {any} value the value to be checked
 *
 * @return {boolean} True if it is, false if it isn't
 */
function notExists(value) {
    return !exists(value);
}
/**
 * A more useful type check than typeof.
 * Guards against gotchas like null and array being an object and NaN being a number.
 * For this purpose it also explicitly states Infinity as a type rather than a number.
 *
 * @param {any} value The value to check the type of.
 *
 * @return {string|null} The type of the value.
 */
function checkType(value) {
    if (value === undefined ||
        value === null ||
        value === Infinity ||
        value === -Infinity) {
        return `${value}`;
    }
    if (isNum(value))
        return 'number';
    if (isStr(value))
        return 'string';
    if (isBool(value))
        return 'boolean';
    if (isFn(value))
        return 'function';
    if (isObj(value))
        return 'object';
    if (Array.isArray(value))
        return 'array';
    if (isNaN(value))
        return `${NaN}`;
    return '';
}

const optionsFormatKeys = ['default', 'ss', 'SS', 'mm', 'MM', 'hh', 'HH', 'dd', 'DD'];
const timeValues = optionsFormatKeys.slice(1);
/**
 * Creates an object using the provided object of functions
 * and assigns it to all or their respective values.
 *
 * @param {Object} formatValues - The object of functions to assign to the time values.
 *
 * @return {Object} - The created object of functions.
 */
function makeValues(formatValues) {
    return timeValues.reduce((obj, item) => ({
        ...obj,
        [item]: exists(formatValues[item]) ? formatValues[item] : exists(formatValues.default) ? formatValues.default : zeroPad
    }), {});
}
/**
 * Builds an options object from default and custom options.
 *
 * @param {Object} [newOptions] - Optional custom options.
 * @param {Object} [oldOptions] - Optional previous options.
 *
 * @throws If any option is invalid.
 *
 * @return {Object} Compiled options from default and custom.
 */
function buildOptions(newOptions, oldOptions) {
    // Run through validation first, than create the options afterwards.
    if (exists(newOptions)) {
        const { formatOutput, countdown, formatValues } = newOptions;
        if (exists(formatOutput) && isNotStr(formatOutput)) {
            throw new TypeError(`Expected formatOutput to be a string; instead got: ${checkType(formatOutput)}`);
        }
        if (exists(countdown) && isNotBool(countdown)) {
            throw new TypeError(`Expected countdown to be a boolean; instead got: ${checkType(countdown)}`);
        }
        if (exists(formatValues)) {
            if (isObj(formatValues)) {
                let toError = false;
                let error = 'Expected formatValues to contain a list of keys with functions that return a string or number; instead got:\n';
                /**
                 * Runs through each key to check that it:
                 *  - Is a valid property, see optionsFormatKeys above.
                 *  - It's value is a function and it returns either a string or number.
                 *
                 * It will then create a string for all errors it finds,
                 * than throw the error after its finished the loop, if any errors are found.
                 */
                Object.keys(formatValues).forEach((key) => {
                    const value = formatValues[key];
                    if (!optionsFormatKeys.includes(key)) {
                        error += ` '${key}': is not a recognised property, should be one of: ${optionsFormatKeys.map(val => ` '${val}'`).toString().trim()}\n`;
                        toError = true;
                    }
                    else if (isFn(value)) {
                        if (isNotNum(value(5)) && isNotStr(value(5))) {
                            error += ` '${key}': the return type for this function is not a string or number, is: ${checkType(value(5))}\n`;
                            toError = true;
                        }
                    }
                    else {
                        error += ` '${key}': is not a function, is: ${checkType(value)}\n`;
                        toError = true;
                    }
                });
                if (toError) {
                    throw new Error(error);
                }
            }
            else {
                throw new TypeError(`Expected formatValues to be a an object of functions; instead got: ${checkType(formatValues)}`);
            }
        }
    }
    // Build formatValues option.
    if (exists(newOptions)) {
        const { formatValues } = newOptions;
        if (exists(formatValues)) {
            let newFormatValues;
            // If oldOptions provided, merge previous formatValues with new ones
            // Otherwise make new ones from newOptions.
            if (exists(oldOptions)) {
                newFormatValues = makeValues({
                    ...oldOptions.formatValues,
                    ...formatValues
                });
            }
            else {
                newFormatValues = makeValues(formatValues);
            }
            newOptions.formatValues = newFormatValues;
        }
    }
    return {
        formatOutput: 'DD hh:{mm:ss}',
        countdown: true,
        formatValues: makeValues({}),
        ...oldOptions,
        ...newOptions
    };
}

/**
 * Converts seconds to time format.
 *
 * @param {Number} seconds The seconds to convert.
 * @param {Object} [options] The options to build the string.
 *
 * Accepts a full options object passed into `create` but only uses `formatOuput` and `formatValues`
 * @param {Boolean} [toBuild=true] Specifies wether to run the buildOptions function or not.
 *
 * For example, if you don't have a fully formed options object to pass in, or none at all.
 * Primarily used in `Timr#formatTime` to avoid making unnecessary calls to buildOptions on every tick.
 *
 * Setting to false will require `options.formatOuput` and `options.formatValues` to exist
 *
 * @return {Object} An object that that contains the `formattedTime` and the
 * `raw` values used to calculate the time.
 */
function formatTime(seconds, options, toBuild = true) {
    /**
     * If options doesn't exist, build regardless.
     * If it does and toBuild = true, run buildOptions
     * Otherwise just return passed in options
     */
    const { formatOutput, formatValues } = (() => {
        if (isObj(options)) {
            if (toBuild || (isNotStr(options.formatOutput) && (isNotFn(options.formatValues)))) {
                return buildOptions(options);
            }
            else {
                return options;
            }
        }
        else {
            return buildOptions();
        }
    })();
    // @ts-expect-error
    const raw = {};
    raw.SS = seconds;
    raw.MM = Math.floor(raw.SS / 60);
    raw.HH = Math.floor(raw.MM / 60);
    raw.DD = Math.floor(raw.HH / 24);
    raw.ss = Math.floor(seconds % 60);
    raw.mm = Math.floor(raw.MM % 60);
    raw.hh = Math.floor(raw.HH % 24);
    raw.dd = raw.DD;
    let stringToFormat = formatOutput;
    const leftBracketPosition = stringToFormat.lastIndexOf('{') + 1;
    const rightBracketPosition = stringToFormat.indexOf('}');
    /**
     * Get values sitting in-between { } (protectedValues)
     * Match returns empty strings for groups it can't find so filter them out.
     */
    const protectedValues = stringToFormat
        .slice(leftBracketPosition, rightBracketPosition)
        .match(/DD|HH|MM|SS/gi);
    // If protectedValues exist, apply logic to removing them
    // otherwise, format string exactly as it appears
    if (protectedValues != null) {
        let lastValueAboveZero;
        if (raw.DD > 0)
            lastValueAboveZero = 'DD';
        else if (raw.HH > 0)
            lastValueAboveZero = 'HH';
        else if (raw.MM > 0)
            lastValueAboveZero = 'MM';
        else
            lastValueAboveZero = 'SS';
        /**
         * If the lastValueAboveZero is inbetween the protectedValues than don't remove them.
         * E.G. starting string: HH:{mm:ss} and lastValueAboveZero is 'ss' than begin slice
         * at leftBracketPosition rather than beginning at ss.
         */
        const tempSlice = stringToFormat.search(new RegExp(lastValueAboveZero, 'ig'));
        const beginSlice = tempSlice >= leftBracketPosition ? leftBracketPosition : tempSlice;
        stringToFormat = stringToFormat.slice(beginSlice >= 0 ? beginSlice : 0);
    }
    // Replaces all values in string with their respective number values
    const formattedTime = stringToFormat
        // Remove any remaining curly braces before formatting.
        .replace(/\{|\}/g, '')
        // Apply specific function form formatValues to value.
        .replace(/DD|HH|MM|SS/gi, (match) => formatValues[match](raw[match]));
    return { formattedTime, raw };
}

/**
 * Converts time format (HH:MM:SS) into seconds.
 *
 * Automatically rounds the returned number to avoid errors
 * with floating point values.
 *
 * If a pattern is provided (25h / 25m / 25d), than
 * it is converted before being validated and processed.
 *
 * @param {string|number} time The time to be converted.
 * If a number is provided it will simply return that number.
 *
 * @throws If the provided time is neither a number nor a string.
 * @throws If the provided time is a negative number.
 * @throws If the provided time is not in the correct format HH:MM:SS.
 *
 * @return {number} The time in seconds.
 */
function timeToSeconds(time) {
    // If a positive number, skip processing and just return the rounded number.
    if (isNum(time) && time >= 0)
        return Math.round(time);
    let newTime = time;
    // Converts '25m', '25h' && '25d' into '25:00', '25:00:00' & '600:00:00' respectivley.
    if (isStr(newTime) && /^\d+[mhd]$/i.test(newTime)) {
        switch (newTime.slice(-1)) {
            case 'm':
                newTime = newTime.replace(/^(\d+)m$/i, '$1:00');
                break;
            case 'h':
                newTime = newTime.replace(/^(\d+)h$/i, '$1:00:00');
                break;
            case 'd':
                newTime = newTime.replace(/^(\d+)d$/i, (m, g1) => `${parseInt(g1) * 24}:00:00`);
                break;
        }
    }
    if (isNotStr(newTime) && isNotNum(newTime)) {
        throw new TypeError(`Expected time to be a string or number, instead got: ${checkType(newTime)}`);
    }
    // Checks if the string can't be converted to a number, i.e someone passed 10:00.
    // Then tests for the correct format
    if (isNaN(Number(newTime)) && isStr(newTime)) {
        if (!/^(\d+)(:\d+)?(:\d+)?$/.test(newTime)) {
            throw new SyntaxError(`Expected time to be in (hh:mm:ss) format, instead got: ${newTime}`);
        }
    }
    if (Number(newTime) < 0) {
        throw new Error(`Time cannot be a negative number, got: ${newTime}`);
    }
    return Math.round(newTime.split(':').reduce((prev, curr, index, arr) => {
        if (arr.length === 3) {
            if (index === 0)
                return prev + (Number(curr) * 60 * 60);
            if (index === 1)
                return prev + (Number(curr) * 60);
        }
        if (arr.length === 2) {
            if (index === 0)
                return prev + (Number(curr) * 60);
        }
        return prev + Number(curr);
    }, 0));
}

/**
 * Confirms if provided startTime is a futureDate parsable by dateToSeconds
 *
 * @param startTime the startTime
 *
 * @return True if it is, false otherwise
 */
function isDateFormat(startTime) {
    return isInstanceOf(startTime, Date) || (isStr(startTime) && /^\d{4}-\d{2}-\d{2}/.test(startTime));
}
/**
 * Converts a date object or date string into seconds until that date/time.
 *
 * @param {string|Date} startTime The date object or date string.
 *
 * Format should be: `((year)-(month)-(day) [[(hour)]:(minute):(second(s))]`.
 * @param {string|Date} backupStartTime startTime to use if provided startTime is in the past.
 *
 * @throws If the date string is not in the correct format.
 * @throws If the date string is in the correct format but can't be parsed, for example by using `13` for the month.
 * @throws If the date is in the past (unless provided `backupStartTime` is not in the past).
 * @throws If the date is neither a `string` or `Date`.
 *
 * @return {Number} Returns the converted seconds.
 */
function dateToSeconds(startTime, backupStartTime) {
    // startTime in MS from epoch.
    let parsedStartTime;
    if (isInstanceOf(startTime, Date)) {
        parsedStartTime = startTime.getTime();
    }
    else if (isStr(startTime)) {
        const error = new Error('The provided date is not in the right format or is incorrect.\n' +
            'Expected a string in the format: YYYY-MM-DD[ HH:MM[:SS]].\n' +
            '(year)-(month)-(day) (hour):(minute):(second(s))\n' +
            'Time is optional and seconds is optional if time provided.\n' +
            `You passed: "${startTime}"`);
        try {
            // @ts-expect-error try catch will catch nullpointer failure
            const [year, month, ...rest] = startTime
                .match(/^(\d{4})-(\d{2})-(\d{2})(?: (\d{2}):(\d{2})(?::(\d{2}))?)?$/i)
                .slice(1)
                .map((value) => parseInt(value))
                .filter(Boolean);
            parsedStartTime = new Date(year, month - 1, ...rest).getTime();
            if (isNaN(parsedStartTime)) {
                throw error;
            }
        }
        catch (err) {
            throw error;
        }
    }
    else {
        throw new Error(`Expected startTime to be a string or Date object, instead got: ${checkType(startTime)}`);
    }
    const dateNow = new Date();
    // startTime in seconds from now, using parsedStartTime above.
    const startTimeInSeconds = Math.ceil((parsedStartTime - dateNow.getTime()) / 1000);
    if (startTimeInSeconds < 0) {
        if (exists(backupStartTime)) {
            return dateToSeconds(backupStartTime);
        }
        throw new Error('When passing a date/time, it cannot be in the past. ' +
            `You passed: "${startTime}". It's currently: "` +
            `${zeroPad(dateNow.getFullYear())}-${zeroPad(dateNow.getMonth() + 1)}-` +
            `${zeroPad(dateNow.getDate())} ` +
            `${zeroPad(dateNow.getHours())}:${zeroPad(dateNow.getMinutes())}:` +
            `${zeroPad(dateNow.getSeconds())}"`);
    }
    return startTimeInSeconds;
}

/**
 * Creates an EventEmitter.
 *
 * This is a super slimmed down version of nodes EventEmitter.
 *
 * This is only intended for internal use, as there is
 * no real error checking.
 */
class EventEmitter {
    constructor() {
        this.events = {};
    }
    /**
     * Registers a listener to an event array.
     *
     * @param event The event to attach to.
     * @param listener The event listener.
     */
    on(event, listener) {
        if (notExists(this.events[event])) {
            this.events[event] = [];
        }
        this.events[event].push(listener);
    }
    /**
     * Emits an event, calling all listeners store
     * against the provided event.
     *
     * @param event The event to emit.
     * @param args The functions to run against the event.
     */
    emit(event, ...args) {
        if (exists(this.events[event])) {
            this.events[event].forEach((listener) => {
                listener.apply(this, args);
            });
        }
    }
    /**
     * Removes all listeners.
     */
    removeAllListeners() {
        this.events = {};
    }
}

/**
 * A Status represents the current state of a Timr object.
 */
var Status;
(function (Status) {
    Status["initialized"] = "initialized";
    Status["started"] = "started";
    Status["paused"] = "paused";
    Status["stopped"] = "stopped";
    Status["finished"] = "finished";
    Status["destroyed"] = "destroyed";
})(Status || (Status = {}));

class Timr extends EventEmitter {
    /**
     * Creates a Timr.
     *
     * @param {String|Number} startTime The starting time for the Timr object.
     * @param {Object} [options] Options to customise the timer.
     *
     * @throws If the provided startTime is neither a number or a string,
     * or, incorrect format.
     */
    constructor(startTime, options) {
        super();
        // options needs to be built before startTime is set,
        // so it can work out the future date properly.
        this.changeOptions(options);
        this.setStartTime(startTime);
        this.status = Status.initialized;
    }
    /**
     * Countdown function.
     *
     * Bound to a setInterval when start() is called.
     */
    _countdown() {
        this.currentTime -= 1;
        this.emit('ticker', {
            ...this.formatTime(),
            percentDone: this.percentDone(),
            currentTime: this.currentTime,
            startTime: this.startTime,
            self: this
        });
        if (this.currentTime <= 0) {
            this.clear();
            this.currentTime = this.startTime;
            this.emit('finish', this);
            this.status = Status.finished;
        }
    }
    /**
     * Stopwatch function.
     *
     * Bound to a setInterval when start() is called.
     */
    _stopwatch() {
        this.currentTime += 1;
        this.emit('ticker', {
            ...this.formatTime(),
            currentTime: this.currentTime,
            startTime: this.startTime,
            self: this
        });
    }
    _listener(name, listener) {
        if (isNotFn(listener)) {
            throw new TypeError(`Expected ${name} to be a function, instead got: ${checkType(listener)}`);
        }
        this.on(name, listener);
        return this;
    }
    /**
     * Starts the Timr.
     *
     * @param {Number} [delay] Optional delay in ms to start the timer.
     *
     * @return {Object} Returns a reference to the Timr so calls can be chained.
     */
    start(delay) {
        if (!this.started()) {
            if (this.options.countdown && this.startTime === 0) {
                throw new Error('Unable to start timer when countdown = true and startTime = 0. ' +
                    'This would cause the timer to count into negative numbers and never stop. ' +
                    'Try setting countdown to false or amending the startTime.');
            }
            const startFn = () => {
                /**
                 * futureDate records the original date used when futureDate option is set to true,
                 * this will re-run setStarTime to ensure the startTime is upto date.
                 *
                 * Note: Inside startFn so that delay works properly, if it was outside this scope,
                 * the startTime would be out of sync after the delay finishes.
                 */
                if (exists(this.futureDate)) {
                    this.setStartTime(this.futureDate);
                }
                this.status = Status.started;
                this.timer = this.options.countdown
                    ? setInterval(() => this._countdown(), 1000)
                    : setInterval(() => this._stopwatch(), 1000);
            };
            if (exists(delay)) {
                if (isNotNum(delay)) {
                    throw new TypeError(`The delay argument passed to start must be a number, you passed: ${checkType(delay)}`);
                }
                this.delayTimer = setTimeout(startFn, delay);
            }
            else {
                startFn();
            }
            this.emit('onStart', this);
        }
        else {
            this.emit('onAlreadyStarted', this);
        }
        return this;
    }
    /**
     * Pauses the Timr.
     *
     * @return {Object} Returns a reference to the Timr so calls can be chained.
     */
    pause() {
        this.clear();
        this.emit('onPause', this);
        this.status = Status.paused;
        return this;
    }
    /**
     * Stops the Timr, resetting the `currentTime` to the `startTime`.
     *
     * @return {Object} Returns a reference to the Timr so calls can be chained.
     */
    stop() {
        this.clear();
        this.currentTime = this.startTime;
        this.emit('onStop', this);
        this.status = Status.stopped;
        return this;
    }
    /**
     * Clears the Timr, clearing the internal timer for both the `delayTimer`
     * and general `timer`.
     *
     * @return {Object} Returns a reference to the Timr so calls can be chained.
     */
    clear() {
        clearInterval(this.timer);
        clearTimeout(this.delayTimer);
        return this;
    }
    /**
     * Destroys the Timr, clearing the intervals (as in `clear()), removing all event listeners and removing,
     * from the store (if it's in one).
     *
     * @return {Object} Returns a reference to the Timr so calls can be chained.
     */
    destroy() {
        this.emit('onDestroy', this);
        this.clear().removeAllListeners();
        // removeFromStore is added when the timr is added to a store,
        // so need to check if it's in a store before removing it.
        if (isFn(this.removeFromStore)) {
            this.removeFromStore();
        }
        this.status = Status.destroyed;
        return this;
    }
    /**
     * Called every second the timer ticks down.
     *
     * @throws If the argument is not of type function.
     *
     * @param {Function} listener Function to added to events.
     * @return {Object} Returns a reference to the Timr so calls can be chained.
     */
    ticker(listener) {
        return this._listener('ticker', listener);
    }
    /**
     * Called once when the timer finishes.
     *
     * This will never be called when using a stopwatch.
     *
     * @throws If the argument is not of type function.
     *
     * @param {Function} listener Function to added to events.
     * @return {Object} Returns a reference to the Timr so calls can be chained.
     */
    finish(listener) {
        return this._listener('finish', listener);
    }
    /**
     * Called when the timer starts.
     *
     * @throws If the argument is not of type function.
     *
     * @param {Function} listener Function to added to events.
     * @return {Object} Returns a reference to the Timr so calls can be chained.
     */
    onStart(listener) {
        return this._listener('onStart', listener);
    }
    /**
    * Called when the timer is already running and start is called
    *
    * @throws If the argument is not of type function.
    *
    * @param {Function} listener Function to added to events.
    * @return {Object} Returns a reference to the Timr so calls can be chained.
    */
    onAlreadyStarted(listener) {
        return this._listener('onAlreadyStarted', listener);
    }
    /**
     * Called when the timer is paused.
     *
     * @throws If the argument is not of type function.
     *
     * @param {Function} listener Function to added to events.
     * @return {Object} Returns a reference to the Timr so calls can be chained.
     */
    onPause(listener) {
        return this._listener('onPause', listener);
    }
    /**
     * Called when the timer is stopped.
     *
     * @throws If the argument is not of type function.
     *
     * @param {Function} listener Function to added to events.
     * @return {Object} Returns a reference to the Timr so calls can be chained.
     */
    onStop(listener) {
        return this._listener('onStop', listener);
    }
    /**
     * Called when the timer is destroyed.
     *
     * @throws If the argument is not of type function.
     *
     * @param {Function} listener Function to added to events.
     * @return {Object} Returns a reference to the Timr so calls can be chained.
     */
    onDestroy(listener) {
        return this._listener('onDestroy', listener);
    }
    /**
     * Converts seconds to time format.
     *
     * This is provided to the ticker function.
     *
     * @param {String} [time=currentTime] optionally format the startTime
     *
     * @return {Object} The `formattedTime` and `raw` values.
     */
    formatTime(time = 'currentTime') {
        return formatTime(this[time], this.options, false);
    }
    /**
     * Returns the time elapsed in percent.
     * This is provided to the ticker.
     *
     * @return {Number} Time elapsed in percent.
     */
    percentDone() {
        return 100 - Math.round((this.currentTime / this.startTime) * 100);
    }
    /**
     * Creates / changes `options` object.
     *
     * Merges with existing or default options.
     *
     * @param {Object} options The options to create / change.
     * @return {Object} Returns a reference to the Timr so calls can be chained.
     */
    changeOptions(options) {
        this.options = buildOptions(options, this.options);
        return this;
    }
    /**
     * Sets new `startTime` after Timr has been created.
     *
     * Will clear `currentTime` and reset to new `startTime`.
     *
     * @param {String|Number} startTime The new `startTime`.
     *
     * @throws If no startTime is provided.
     * @throws If the provided time is not a string.
     * @throws If the provided time is not in the correct format HH:MM:SS.
     * @throws If the date string is not in the correct format.
     * @throws If the date string is in the correct format but can't be parsed, for example by using `13` for the month.
     * @throws If the date is in the past (unless provided `backupStartTime` is not in the past).
     *
     * @return {Object} Returns a reference to the Timr so calls can be chained.
     */
    setStartTime(startTime, backupStartTime) {
        this.clear();
        if (notExists(startTime)) {
            throw new Error('You must provide a startTime value.');
        }
        let newStartTime;
        if (isDateFormat(startTime)) {
            newStartTime = dateToSeconds(startTime, backupStartTime ?? this.options.backupStartTime);
            this.futureDate = startTime;
        }
        else {
            newStartTime = timeToSeconds(startTime);
            this.futureDate = null;
        }
        this.startTime = newStartTime;
        this.currentTime = newStartTime;
        return this;
    }
    /**
     * Shorthand for `this.formatTime(time).formattedTime`
     */
    getFt(time = 'currentTime') {
        return this.formatTime(time).formattedTime;
    }
    /**
     * Shorthand for `this.formatTime(time).raw`
     */
    getRaw(time = 'currentTime') {
        return this.formatTime(time).raw;
    }
    /**
     * Gets the Timrs `startTime`.
     *
     * @return {Number} Start time in seconds.
     */
    getStartTime() {
        return this.startTime;
    }
    /**
     * Gets the Timrs currentTime.
     *
     * @return {Number} Current time in seconds.
     */
    getCurrentTime() {
        return this.currentTime;
    }
    /**
     * Returns the current status. Or returns a boolean comparing
     * the provided statusName to the current status.
     *
     * @param {Object} [statusName] Optional `Status` to check
     *
     * @return Either the current status, or a boolean confirming the current status if
     * a statusName is provided. Will check `statusName === this.status`
     */
    getStatus(statusName) {
        return exists(statusName)
            ? this.status === statusName
            : this.status;
    }
    /**
     * Returns true if the Timr has started
     *
     * @deprecated please use `this.started()` instead
     *
     * @return {Boolean} True if running, false if not.
     */
    isRunning() {
        return this.started();
    }
    /**
     * Checks whether the timer has been started or not
     *
     * @return {Boolean} True if running, false if not.
     */
    started() {
        return this.getStatus(Status.started);
    }
}

/**
 * Creates a store that can hold multiple Timr objects
 * and perform functions on all of them.
 *
 * @param {Array} [args] - Optional timers to start the store with.
 *
 * @throws If the provided Timr is not a Timr object.
 * @throws If the provided Timr is already in a store.
 *
 * @return {Object} Returns a store object with methods.
 */
function createStore(...args) {
    let timrs = [];
    /**
     * Removes the provided Timr from the store
     *
     * @param {Object} timr The Timr to remove
     */
    function removeFromStore(timr) {
        // Instanceof check required as it's exposed as a store method.
        if (isInstanceOf(timr, Timr)) {
            timrs = timrs.filter(x => x !== timr);
            timr.removeFromStore = null;
        }
    }
    /**
     * Adds the provided Timr to the store.
     *
     * @param {Object} timr A Timr object.
     *
     * @throws If the provided Timr is not a Timr object.
     * @throws If the provided timer is already in a store.
     *
     * @return {Object} The provided Timr object.
     */
    function add(timr) {
        if (isInstanceOf(timr, Timr) && isNotFn(timr.removeFromStore)) {
            timrs.push(timr);
            timr.removeFromStore = () => {
                removeFromStore(timr);
            };
        }
        else {
            throw new Error('Unable to add to store; provided argument is either already in a store ' +
                'or not a timr object.');
        }
        return timr;
    }
    args.forEach(add);
    return {
        add,
        /**
         * Returns an array of all the timers in the store.
         *
         * @return {Array} The array of Timrs.
         */
        getAll() {
            return timrs;
        },
        /**
         * Starts all the timers in the store.
         */
        startAll() {
            timrs.forEach(timr => timr.start());
        },
        /**
         * Pauses all the timers in the store.
         */
        pauseAll() {
            timrs.forEach(timr => timr.pause());
        },
        /**
         * Stops all the timers in the store.
         */
        stopAll() {
            timrs.forEach(timr => timr.stop());
        },
        /**
         * Returns an array of all the timers that are in the same status
         * as the one provided.
         *
         * @param {string} statusName The status to check.
         *
         * @return {Array} An array of timers that are in a state matched by the provided `statusName`
         */
        getStatus(statusName) {
            return timrs.filter(timr => timr.getStatus(statusName));
        },
        /**
         * Returns an array of timers that are currently running.
         *
         * @deprecated Please use `started()` instead
         *
         * @return {Array} The array of running Timrs
         */
        isRunning() {
            return timrs.filter(timr => timr.isRunning());
        },
        /**
         * Returns an array of timers that have started.
         *
         * @return {Array} The array of Timrs that have started
         */
        started() {
            return timrs.filter(timr => timr.started());
        },
        removeFromStore,
        /**
         * Destroys all Timrs in the store
         */
        destroyAll() {
            timrs.forEach(timr => timr.destroy());
            timrs = [];
        }
    };
}

/**
 * Creates a new Timr object.
 *
 * @param {string|number|Date|object} startTime The starting time. Or an object with startTime and options.
 *
 * `string` Examples of accepted format:
 *  - `'10:00:00'` A 10 hour timer.
 *  - `'10:00'` A 10 minute timer.
 *  - `'10'` A 10 second timer.
 *  - `'10m'` A 10 minute timer.
 *  - `'10h'` A 10 hour timer.
 *  - `'10d'` A 10 day timer.
 *  - `'2021-12-25'` Creates a countdown timer to midnight on Christmas Day.
 *  - `'2021-12-25 10:00'` Creates a countdown timer to 10am on Christmas Day.
 *
 * `number` In seconds.
 *
 * `Date` The timer will countdown to the point in time specified by the date.
 *
 * `object` Options to customise the timer with a `startTime` property. See below.
 *
 * **Note**: If an `object` is provided to `startTime` the `options` argument will be ignored.
 *
 * @param {Object} [options] Options to customise the timer.
 *
 * Accepts the following properties, see `README.md` for further details:
 * - `formatOutput` How the time should be formatted as a string.
 * - `countdown` Whether the timer should count down or up.
 * - `formatValues` An object of functions to format each part of the formatted string individually.
 * - `backupStartTime` A backupStartTime for timers that countdown to a futureDate, in case that time is in the past.
 *
 * @return {Object} A new Timr object.
 */
function create(startTime, options) {
    if (isObj(startTime)) {
        if (exists(startTime.startTime)) {
            return new Timr(startTime.startTime, startTime);
        }
        throw new Error('When providing only an object when creating a timer, it must have a startTime property.');
    }
    return new Timr(startTime, options);
}

export { Status, create, createStore, dateToSeconds, formatTime, timeToSeconds, zeroPad };
