/// <reference types="node" />
import EventEmitter from './EventEmitter';
import { FormattedTime, Listener, OptionalOptions, Options, Raw } from './types/common';
import { Status } from './types/enums';
import { ITimr } from './types/ITimr';
declare class Timr extends EventEmitter implements ITimr {
    timer: NodeJS.Timeout;
    delayTimer: NodeJS.Timeout;
    currentTime: number;
    startTime: number;
    options: Options;
    futureDate: string | number | Date | null;
    status: Status;
    removeFromStore?: null | (() => void);
    [key: string]: any;
    /**
     * Creates a Timr.
     *
     * @param {String|Number} startTime The starting time for the Timr object.
     * @param {Object} [options] Options to customise the timer.
     *
     * @throws If the provided startTime is neither a number or a string,
     * or, incorrect format.
     */
    constructor(startTime: string | number | Date, options?: OptionalOptions);
    /**
     * Countdown function.
     *
     * Bound to a setInterval when start() is called.
     */
    private _countdown;
    /**
     * Stopwatch function.
     *
     * Bound to a setInterval when start() is called.
     */
    private _stopwatch;
    private _listener;
    /**
     * Starts the Timr.
     *
     * @param {Number} [delay] Optional delay in ms to start the timer.
     *
     * @return {Object} Returns a reference to the Timr so calls can be chained.
     */
    start(delay?: number): this;
    /**
     * Pauses the Timr.
     *
     * @return {Object} Returns a reference to the Timr so calls can be chained.
     */
    pause(): this;
    /**
     * Stops the Timr, resetting the `currentTime` to the `startTime`.
     *
     * @return {Object} Returns a reference to the Timr so calls can be chained.
     */
    stop(): this;
    /**
     * Clears the Timr, clearing the internal timer for both the `delayTimer`
     * and general `timer`.
     *
     * @return {Object} Returns a reference to the Timr so calls can be chained.
     */
    clear(): this;
    /**
     * Destroys the Timr, clearing the intervals (as in `clear()), removing all event listeners and removing,
     * from the store (if it's in one).
     *
     * @return {Object} Returns a reference to the Timr so calls can be chained.
     */
    destroy(): this;
    /**
     * Called every second the timer ticks down.
     *
     * @throws If the argument is not of type function.
     *
     * @param {Function} listener Function to added to events.
     * @return {Object} Returns a reference to the Timr so calls can be chained.
     */
    ticker(listener: Listener): this;
    /**
     * Called once when the timer finishes.
     *
     * This will never be called when using a stopwatch.
     *
     * @throws If the argument is not of type function.
     *
     * @param {Function} listener Function to added to events.
     * @return {Object} Returns a reference to the Timr so calls can be chained.
     */
    finish(listener: Listener): this;
    /**
     * Called when the timer starts.
     *
     * @throws If the argument is not of type function.
     *
     * @param {Function} listener Function to added to events.
     * @return {Object} Returns a reference to the Timr so calls can be chained.
     */
    onStart(listener: Listener): this;
    /**
    * Called when the timer is already running and start is called
    *
    * @throws If the argument is not of type function.
    *
    * @param {Function} listener Function to added to events.
    * @return {Object} Returns a reference to the Timr so calls can be chained.
    */
    onAlreadyStarted(listener: Listener): this;
    /**
     * Called when the timer is paused.
     *
     * @throws If the argument is not of type function.
     *
     * @param {Function} listener Function to added to events.
     * @return {Object} Returns a reference to the Timr so calls can be chained.
     */
    onPause(listener: Listener): this;
    /**
     * Called when the timer is stopped.
     *
     * @throws If the argument is not of type function.
     *
     * @param {Function} listener Function to added to events.
     * @return {Object} Returns a reference to the Timr so calls can be chained.
     */
    onStop(listener: Listener): this;
    /**
     * Called when the timer is destroyed.
     *
     * @throws If the argument is not of type function.
     *
     * @param {Function} listener Function to added to events.
     * @return {Object} Returns a reference to the Timr so calls can be chained.
     */
    onDestroy(listener: Listener): this;
    /**
     * Converts seconds to time format.
     *
     * This is provided to the ticker function.
     *
     * @param {String} [time=currentTime] optionally format the startTime
     *
     * @return {Object} The `formattedTime` and `raw` values.
     */
    formatTime(time?: 'currentTime' | 'startTime'): FormattedTime;
    /**
     * Returns the time elapsed in percent.
     * This is provided to the ticker.
     *
     * @return {Number} Time elapsed in percent.
     */
    percentDone(): number;
    /**
     * Creates / changes `options` object.
     *
     * Merges with existing or default options.
     *
     * @param {Object} options The options to create / change.
     * @return {Object} Returns a reference to the Timr so calls can be chained.
     */
    changeOptions(options?: OptionalOptions): this;
    /**
     * Sets new `startTime` after Timr has been created.
     *
     * Will clear `currentTime` and reset to new `startTime`.
     *
     * @param {String|Number} startTime The new `startTime`.
     *
     * @throws If no startTime is provided.
     * @throws If the provided time is not a string.
     * @throws If the provided time is not in the correct format HH:MM:SS.
     * @throws If the date string is not in the correct format.
     * @throws If the date string is in the correct format but can't be parsed, for example by using `13` for the month.
     * @throws If the date is in the past (unless provided `backupStartTime` is not in the past).
     *
     * @return {Object} Returns a reference to the Timr so calls can be chained.
     */
    setStartTime(startTime: string | number | Date, backupStartTime?: string | Date): this;
    /**
     * Shorthand for `this.formatTime(time).formattedTime`
     */
    getFt(time?: 'currentTime' | 'startTime'): string;
    /**
     * Shorthand for `this.formatTime(time).raw`
     */
    getRaw(time?: 'currentTime' | 'startTime'): Raw;
    /**
     * Gets the Timrs `startTime`.
     *
     * @return {Number} Start time in seconds.
     */
    getStartTime(): number;
    /**
     * Gets the Timrs currentTime.
     *
     * @return {Number} Current time in seconds.
     */
    getCurrentTime(): number;
    /**
     * Returns the current status. Or returns a boolean comparing
     * the provided statusName to the current status.
     *
     * @param {Object} [statusName] Optional `Status` to check
     *
     * @return Either the current status, or a boolean confirming the current status if
     * a statusName is provided. Will check `statusName === this.status`
     */
    getStatus(statusName?: Status): boolean | Status;
    /**
     * Returns true if the Timr has started
     *
     * @deprecated please use `this.started()` instead
     *
     * @return {Boolean} True if running, false if not.
     */
    isRunning(): boolean;
    /**
     * Checks whether the timer has been started or not
     *
     * @return {Boolean} True if running, false if not.
     */
    started(): boolean;
}
export default Timr;
